var mysql = require('mysql');
var logger = require('log4js').getLogger('io');
var DatabaseService = require('../database-service');

var pattern = {
    variableAndIdentifier: /(@(\d+-)?\w+(\.\w+)*@)|(\$(\d+-)?\w+(\.\w+)*\$)/g,
    variable: /@(\d+-)?\w+(\.\w+)*@/g,
    clientVariable: /(@\d+-\w+(\.\w+)*@)|(\$\d+-\w+(\.\w+)*\$)/g,
    identifier:  /\$(\d+-)?\w+(\.\w+)*\$/g
};

var MESSAGE = {
    en: {
        BEFORE_SQL_SYNTAX_ERR: 'The before sql function has syntax err. Reason: ',
        SQL_COMPILE_ERR: 'Error occured when compile sql. Reason: ',
        PROCEDURE_SYNTAX_ERR: 'Error occured when compile procedure. Reason: ',
        PROCEDURE_EXE_ERR: 'Error occured when executing procedure. Reason: ',
        TIMEOUT: 'Exectue timeout. Please check if the done() method is called correctly',
        OP_SUCC: 'Operation successfully!',
        OP_FAILED: 'Operation failed'
    },
    'zh-cn': {
        BEFORE_SQL_SYNTAX_ERR: 'SQL 前置执行函数有语法错误，请检查书写是否正确。原因：',
        SQL_COMPILE_ERR: '解析 SQL 时发生错误，请检查书写是否正确。原因：',
        PROCEDURE_SYNTAX_ERR: '代码有语法错误，请检查书写是否正确。原因：',
        PROCEDURE_EXE_ERR: '执行过程时发生错误，原因: ',
        TIMEOUT: '执行超时。请检查 done 方法是否被正确地调用，确保后续程序执行。',
        OP_SUCC: '操作成功！',
        OP_FAILED: 'Operation failed'
    }
};

exports.before = function(req, res, next) {
    var dbConfiguration;
    try {
        dbConfiguration = JSON.parse(req.query.database);
    } catch (e) {
        return next(
            new Error('Invalid database setting for preview.')
        );
    }

    var criteria;
    try {
        criteria = JSON.parse(req.query.criteria);
    } catch (e) {
        return res.end('Invalid criteria');
    }
    if (!criteria.query) {
        return res.end();
    }

    // Mock server vars for user debug
    var serverVars = req.query.serverVars;
    try {
        serverVars = JSON.parse(decodeURIComponent(serverVars));
    } catch(e) {
        serverVars = {};
    }
    criteria.serverVars = serverVars;
    req.criteria = criteria;
    req.dbService = new DatabaseService(dbConfiguration);
    next();
};

exports.queryData = function(req, res, next) {
    req.dbService.criteriaQuery(req.criteria, function(err, result) {
        if (err) {
            res.jsonp({
                success: false,
                message: err.message + ' CODE: ' + err.code
            });
            return;
        }
        res.jsonp({
            success: true,
            data: result
        });
    });
};

exports.exportData = function(req, res, next) {
    var format = req.criteria.exportFormat || "xlsx";
    var formatter = DataFormatter[ format ];
    if (typeof formatter !== 'function') {
        return next(new Error('Invalid format: ' + format));
    }
    req.dbService.criteriaQuery(req.criteria, function(err, result) {
        if (err) {
            return next(err);
        }
        result = formatter(result.rows, req.criteria.colModel);
        res.set({
            "Content-Type": "application/octet-stream",
            "Content-Disposition": "attachment; filename=" 
              + (new Date()).getTime() + "." + format
        });
        res.send(result);
    });
};

var xlsx = require( "node-xlsx" );
var fs = require( "fs" );
var DataFormatter = {
    xlsx: function( result, colModel ) {
        colModel = colModel || [];
        var header = [];
        // generate header;
        for ( var i = 0; i < colModel.length; i++ ) {
            header.push( colModel[ i ].name );
        }
        // generate data;
        var data = [ header ], row;
        for ( var i = 0; i < result.length; i++ ) {
            r = result[ i ];
            row = [];
            var n = 0, v;
            for ( var j in r ) {
                v = r[j];
                
                if ( ( j === "parse" || j === "_typeCast") 
                    && typeof v === "function" ) {
                    break;
                }
                row.push(v);
                // row.push( {
                //     value: v === null ? '' : v,
                //     formatCode: ( colModel[n] || {} ).formatCode
                // } );
                n++;
            }
            data.push( row );
        }
        return xlsx.build( [ {
                "name": "sheet", 
                "data": data
            } ] );
    },
    csv: function(result, colModel) {
        return result;
    },
    json: function( result, colModel ) {
        return result;
    },
    xml: function( result, colModel ) {
        return result;
    }
};

var ESqlCompiler = require('../esql-compiler');

exports.execute = function(req, res, next) {
    var dbConfiguration;
    try {
        dbConfiguration = JSON.parse(req.query.database);
    } catch (e) {
        return next(
            new Error('Invalid database setting for preview.')
        );
    }

    var MSG = MESSAGE[req.cookies.lang || 'zh-cn'];
    var params, procedure;
    try {
        params = JSON.parse( req.query.params );
        procedure = JSON.parse( req.query.procedure );
    } catch (e) {
        return next(new Error('Invalid params or procedure'));
    }

    var params2 = {};
    for (var i in params) {
        params2[i.toUpperCase()] = params[i];
    }
    params = params2;

    // Mock server vars for user debug
    var serverVars = req.query.serverVars;
    try {
        serverVars = JSON.parse(decodeURIComponent(serverVars));
    } catch(e) {
        serverVars = {};
    }

    var f;
    procedure.type = procedure.type || 'sql';
    if (procedure.type === 'sql') {
        var beforeSql, compiledSqlFunc;
        try {
            beforeSql = procedure.beforeSql || 'done()';
            beforeSql = beforeSql.replace( pattern.variable, function( s ) {
                return /\@\w+(\.\w+)*\@/.test(s) 
                    ? 'serverVariables["' + s.replace(/\@/g, '').toUpperCase() + '"]'
                    : 'parameters["' + s.replace(/\@/g, '').toUpperCase() + '"]';
            });
            beforeSql = 'function(parameters, serverVariables, Enhancer, done) {\n'
                        + beforeSql + '\n}';

            beforeSql = eval( "(" + beforeSql + ")" );
        } catch ( err ) {
            return res.jsonp({
                success: false,
                message: MSG.BEFORE_SQL_SYNTAX_ERR + err.message,
                debugInfo: JSON.stringify(err)
            });
        }
        try {
            compiledSqlFunc = ESqlCompiler.compile(procedure.func);
        } catch (err) {
            return res.jsonp({
                success: false,
                message: MSG.SQL_COMPILE_ERR + err.message,
                debugInfo: err.message
            });
        }
        f = function(parameters, serverVariables, Enhancer, done) {
            beforeSql(parameters, serverVariables, Enhancer, function(err, result) {
                if (err) {
                    return done(err);
                }
                if (result && !result.success) {
                    return done(null, result);
                }
                compiledSqlFunc(parameters, serverVariables, Enhancer, done);
            });
        }

    } else if (procedure.type === 'proc') {
        var funcStr;
        try {
            funcStr = procedure.func.replace( pattern.variable, function( s ) {
                return /\@\w+(\.\w+)*\@/.test(s) 
                    ? 'serverVariables["' + s.replace(/\@/g, '').toUpperCase() + '"]'
                    : 'parameters["' + s.replace(/\@/g, '').toUpperCase() + '"]';
            });
            funcStr = 'function(parameters, serverVariables, Enhancer, done) {\n'
                        + funcStr + '\n}';

            f = eval( "(" + funcStr + ")" );
        } catch ( err ) {
            return res.jsonp({
                success: false,
                message: MSG.PROCEDURE_SYNTAX_ERR + err.message,
                debugInfo: JSON.stringify(err)
            });
        }
    }

    // Set timer
    var completed = false;
    try {
        // Mock enhancer;
        var enhancer = {
            getDatabaseService: function() {
                return new DatabaseService(dbConfiguration);
            },
            getDatabaseConfig: function() {
                return dbConfiguration;
            }
            // getSession: function() {

            // },
            // getHttpRequest: function() {

            // },
            // getHttpResponse: function() {

            // }
        };

        setTimeout(function() {
            if (completed) {
                return;
            }
            completed = true;
            res.jsonp({
                success: false,
                message: MSG.TIMEOUT
            });
        }, 3000);
        var t1 = new Date().getTime();

        f( params, serverVars, enhancer, function(err, result) {
            
            if (req.conn) {
                req.conn.destroy();
            }
            if (completed) {
                return;
            }
            completed = true;
            var t2 = new Date().getTime();
            console.debug('PROCEDURE COST: ', (t2 - t1) + 'ms' );
            if (err) {
                // Find error tips
                var errorMessage;
                for (i in procedure.errMsg) {
                    if (err.code === i || ~err.message.indexOf(i)) {
                        errorMessage = procedure.errMsg[i];
                        break;
                    }
                }
                errorMessage = errorMessage
                    || (procedure.errMsg.default || MSG.PROCEDURE_EXE_ERR)
                        + 'Reason: ' + err.message;

                res.jsonp({
                    success: false,
                    message: errorMessage,
                    errorCode: err.code,
                    debugInfo: err.debugInfo
                });
                return;
            }

            result.message = result.message 
                || (result.success ? procedure.succMsg : procedure.errMsg.default)
                || (result.success ? MSG.OP_SUCC : MSG.OP_FAILED);

            res.jsonp(result);
        } );
    } catch ( err ) {
        if (completed) {
            return;
        }
        completed = true;
        return res.jsonp({
            success: false,
            message: MSG.PROCEDURE_EXE_ERR + err.message,
            errInfo: JSON.stringify(err)
        });
    }
};
