var express = require( 'express' );
var router = express.Router();
var projectHost = require('../config').projectHost;
var request = require('request');
var requireNocache = require('require-nocache')(module);
var fs = require('fs-extra');
var path = require('path');

var DatabaseService = require('../database-service');

var customInterfacePathBase = '../repository/project/custom-interface/';

router.all('/custom-interface/p/:project_id/call/:interface_name', function(req, res, next) {

    // Mock getDatabaseService
    req.getDatabaseService = function() {
        var dbConfiguration;
        try {
            if (!req.query._d_b_) {
                return res.status(500).send('Invalid database config.');
            }
            dbConfiguration = JSON.parse(req.query._d_b_);
        } catch (e) {
            return next(
                new Error('Invalid database setting for preview.')
            );
        }
        return new DatabaseService(dbConfiguration);
    };
    
    var serverVars = req.query._s_v_;
    try {
        serverVars = JSON.parse(decodeURIComponent(serverVars));
    } catch(e) {
        serverVars = {};
    }

    // Mock getServerVariables
    req.getServerVariables = function() {
        return serverVars;
    };

    // Mock getUser
    req.getUser = function() {
        return {
            getId: function() {
                return serverVars['USER_ID'];
            },
            getName: function() {
                return serverVars['USER_NAME'];
            },
            getRoles: function() {
                return serverVars['ROLES'];
            }
        }
    };

    res.setHeader('Access-Control-Allow-Origin', '*');

    var pid = req.params.project_id;
    var iname = req.params.interface_name;
    var url = projectHost + '/proj/' + pid + '/httpinterface/name/' + iname;
    var that = this;

    // TODO: install dependencies

    // Get interface definition.
    request({
        url: url,
        headers: {
            'Cookie': 'token=' + req.query._u_t_k_
        }
    }, function(err, r, body) {
        if (err) {
            console.error(err);
            res.status(500)
                .send('Failed to get interface by name ' + iname + '. Reason: ' + err.message);
            return ;
        }
        if (r.statusCode !== 200) {
            console.log(r.statusCode, r.body);
            res.status(500)
                .send('Failed to get interface ' + iname + '. statusCode: ' + r.statusCode);
            return;
        }
        var data, interface;
        try {
            data = JSON.parse(body);
            if (!data.success || !data.result) {
                return res.status(500).send('Failed to get interface ' + iname);
            }
        } catch(err) {
            console.log(err);
            return res.status(403).send('Not authorized for ' + iname);
        }

        if (req.method.toUpperCase() !== data.result.method_type.toUpperCase()) {
            return res.status(404).send('Not found');
        }

        interface = data.result.interface;
        var file = path.resolve(__dirname, customInterfacePathBase + iname);
        fs.ensureFileSync(file);
        fs.writeFileSync(file, interface);

        var f = requireNocache(customInterfacePathBase + iname);

        if (typeof f !== 'function') {
            return res.status(500)
                .send('[Interface Error] No function module is exported.');
        }

        var isCompleted = false;
        res.on('finish', function(e) {
            isCompleted = true;
        });
        setTimeout(function() {
            if (!isCompleted) {
                isCompleted = true;
                res.status(500)
                    .end('[Timeout Error] Interface response timeout, please check if it responses correctly.');
            }
        }, 3500);
        try {
            f.apply(that, [req, res, next]);
        } catch(err) {
            console.log(err);
            isCompleted = true;
            return res.status(500)
                .send('[Interface Calling Error] Error occured when calling interface: '
                    + err.stack);
        }
    });

});

module.exports = router;